#!/usr/bin/env python

"""
Produce syspython code from an inspected program.

Copyright (C) 2006, 2007, 2010, 2011, 2012, 2013 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from micropython.errors import *
from os.path import exists, extsep, join
from micropython.stdcompiler import compiler
import sys
import os

try:
    set
except NameError:
    from sets import Set as set

# Convenience definitions.

special_name = compiler.ast.Name

def quoted_name(s):
    return compiler.ast.Const(s)

def quoted_ref(obj):
    return compiler.ast.CallFunc("static", [quoted_name(obj.full_name())])

def module_attribute(module_name, attrname):
    return special_name(module_name + "." + attrname)

def constant_attribute(parent, attrname):
    return compiler.ast.CallFunc("static", [quoted_name(parent.full_name() + "." + attrname)])

# Special function names.
# Some of the assignment operations cannot be supported unless attribute usage
# observations are being made.

assattr_functions = ("storeattrcontext", "storeattrcontext",
                     "storeattr", "storeattrindexcontextcond",
                     None)
getattr_functions = ("loadattrcontext", "loadattrcontextcond",
                     "loadattr", "loadattrindexcontextcond",
                     "loadconstant")

# Source code classes.

class ConvertedSource(ASTVisitor):

    "A conversion of module source code to syspython."

    def __init__(self, module, program):
        self.visitor = self
        self.module = module
        self.program = program
        self.objtable = program.get_object_table()
        self.current_definition = None
        self.units = []

    def get_unit(self):
        return self.units[-1]

    def get_module(self):
        return self.units[0]

    def to_stream(self, stream):

        "Write the converted code to the given 'stream'."

        module = self.dispatch(self.module.astnode)
        stream.write(str(module))

    def store_value(self, unit, scope, name, value):

        """
        In the given 'unit' and for the given 'scope', store for the given
        'name' the given 'value'.
        """

        if scope == "local":

            # Function locals are stored using a function.

            if isinstance(unit, Function):
                return compiler.ast.CallFunc(
                    special_name("storelocal"),
                    [special_name(name), value]
                    )

            # Class locals are class attribute references.

            elif isinstance(unit, Class):
                return compiler.ast.CallFunc(
                    special_name("storeattrcontext"),
                    [quoted_ref(unit), special_name(name), value]
                    )

            # Module locals are module attribute references.

            elif isinstance(unit, Module):
                return compiler.ast.CallFunc(
                    special_name("storeattr"),
                    [quoted_ref(unit), special_name(name), value]
                    )
            else:
                raise TranslateError("Program unit has no local %r." % name)

        elif scope == "global":

            # Globals are references to module attributes.

            return compiler.ast.CallFunc(
                special_name("storeattr"),
                [quoted_ref(self.get_module()), special_name(name), value]
                )

        elif scope == "builtin":

            # Builtins are accessed via the __builtins__ module.

            return compiler.ast.CallFunc(
                special_name("storeattr"),
                [special_name("__builtins__"), special_name(name), value]
                )

        else:
            # NOTE: This may happen because a class attribute is optimised away.
            return compiler.ast.CallFunc(
                special_name("storeunknown"),
                [special_name(name), value]
                )

    def NOP(self, node):
        return node

    def visitModule(self, node):
        module = node.unit
        self.units.append(module)

        definitions = self.process_definitions(node)
        self.current_definition = None

        # keywords(name, ...)

        keywords = module.keyword_names and [
            compiler.ast.CallFunc(
                special_name("keywords"),
                [special_name(name) for name in module.keyword_names]
                )
            ] or []

        # globalnames(name, ...)

        globalnames = module.module_attribute_names() and [
            compiler.ast.CallFunc(
                special_name("globalnames"),
                [special_name(attr.name) for attr in module.attributes_as_list()]
                )
            ] or []

        # def __main__():
        #     ...

        main = compiler.ast.Function(
            [], "__main__", [], [], 0, "Module initialisation.",
            compiler.ast.Stmt(globalnames + self.dispatch(node.node).nodes)
            )

        self.units.pop()

        return compiler.ast.Module(node.doc, compiler.ast.Stmt(keywords + definitions + [main]))

    # Statements.

    def visitAssert(self, node):
        return compiler.ast.Assert(self.dispatch(node.test), node.fail and self.dispatch(node.fail))

    def visitAssign(self, node):
        expr = self.dispatch(node.expr)
        return compiler.ast.Stmt(
            [compiler.ast.Assign([compiler.ast.AssName("expr", "OP_ASSIGN")], expr)] +
            [self.dispatch(n, compiler.ast.Name("expr")) for n in node.nodes]
            )

    def visitAugAssign(self, node):

        #    lvalue = op(lvalue, expr)
        # -> fn(lvalue, op(lvalue, expr))

        op_name = operator_functions[node.op]

        return self.dispatch(node.node, compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("operator", op_name),
                self.dispatch(node.node), self.dispatch(node.expr)]
            ))

    visitBreak = NOP

    def visitClass(self, node):
        if not used_by_unit(node):
            return compiler.ast.Stmt([])

        self.units.append(node.unit)
        try:
            # Incorporate class body code in the main function.

            if not self.processing_definition(node):
                return self.dispatch(node.code)
            else:
                return self._visitClassDefinition(node)

        finally:
            self.units.pop()

    def _visitClassDefinition(self, node):
        cls = node.unit

        # instattrs(name, ...)
        # clsattrs(name, ...)

        instattrs = cls.instance_attribute_names() and [
            compiler.ast.CallFunc(
                special_name("instattrs"),
                [special_name(attr.name) for attr in cls.instance_attributes_as_list()]
                )
            ] or []

        clsattrs = cls.class_attribute_names() and [
            compiler.ast.CallFunc(
                special_name("clsattrs"),
                [special_name(attr.name) for attr in cls.attributes_as_list()]
                )
            ] or []

        # inherited(superclass, name, ...)
        # ...

        attrs_by_cls = {}
        for attrname, attr in cls.all_class_attributes().items():
            supercls = attr.parent
            if supercls is cls:
                continue
            if not attrs_by_cls.has_key(supercls):
                attrs_by_cls[supercls] = []
            attrs_by_cls[supercls].append(attrname)

        inherited = []

        for supercls, attrnames in attrs_by_cls.items():
            inherited.append(
                compiler.ast.CallFunc(
                    special_name("inherited"),
                    [quoted_ref(supercls)] + [special_name(name) for name in attrnames]
                    ))

        # descendants(name, ...)

        descendants = cls.all_descendants() and [
            compiler.ast.CallFunc(
                special_name("descendants"),
                [special_name(name) for name in cls.all_descendants().keys()]
                )
            ] or []

        # Process all the definitions defined inside the class.

        definitions = self.process_definitions(node)

        return compiler.ast.Class(node.name, [], node.doc,
            compiler.ast.Stmt(instattrs + clsattrs + inherited + descendants + definitions)
            )

    visitContinue = NOP

    def visitDiscard(self, node):
        return compiler.ast.Discard(self.dispatch(node.expr))

    def visitFor(self, node):

        """
        Convert from...

        for <assign> in <list>:
            <body>
        [ else:
            <else_> ]

        ...to...

        _it = iter(<list>)
        while True:
            try:
                <assign> = _it.next()
            except StopIteration:
                [ <else_> ]
                break
            else:
                <body>
        """

        unit = self.get_unit()
        temp = quoted_name(unit.temp_usage)
        unit.temp_usage += 1

        else_nodes = node.else_ and self.dispatch(node.else_).nodes or []

        return compiler.ast.Stmt([
            # storetemp(_it, __builtins__.iter(<list>))
            compiler.ast.CallFunc(special_name("storetemp"), [
                temp,
                compiler.ast.CallFunc(
                    special_name("apply"),
                    [module_attribute("__builtins__", "iter"), self.dispatch(node.list)]
                    )
                ]),
            # while True: ...
            compiler.ast.While(
                special_name("True"),
                # try: ...
                compiler.ast.TryExcept(
                    compiler.ast.Stmt([
                        # <assign> = ...
                        self.dispatch(node.assign,
                            # _it.next()
                            compiler.ast.CallFunc(
                                compiler.ast.CallFunc(special_name("loadattrindex"), [
                                    compiler.ast.CallFunc(special_name("loadtemp"), [temp]),
                                    special_name("next")
                                    ]),
                                []
                                )
                            )
                        ]),
                    # except StopIteration: ...
                    [(special_name("StopIteration"), None, compiler.ast.Stmt(else_nodes + [compiler.ast.Break()]))],
                    # else: ...
                    self.dispatch(node.body)
                    ),
                None
                )
            ])

    def visitFrom(self, node):

        # Generate __main__ function calls for each step in the imported module
        # hierarchy.

        statements = []

        for modname in self.module.get_module_paths(node.modname):
            statements.append(
                compiler.ast.CallFunc(special_name("%s.__main__" % modname ), [])
                )

        for name, alias in node.names:
            statements.append(
                compiler.ast.CallFunc(
                    special_name("storelocal"),
                    [special_name(alias or name),
                    compiler.ast.CallFunc(
                        special_name("loadattr"),
                        [special_name(node.modname), special_name(name)]
                        )
                    ])
                )

        return compiler.ast.Stmt(statements)

    def visitFunction(self, node):
        if not used_by_unit(node):
            return compiler.ast.Stmt([])

        self.units.append(node.unit)

        try:
            if not self.processing_definition(node):

                # Generate rebindings of functions where multiple definitions
                # exist within a scope. Also generate dynamic function object
                # initialisation.

                fn = node.unit
                if fn.name == fn.original_name and not fn.is_dynamic():
                    return compiler.ast.Stmt([])
                else:
                    if fn.is_dynamic():
                        ref = compiler.ast.CallFunc(
                            special_name("makedynamic"),
                            [quoted_ref(fn)] + fn.defaults
                            )
                    else:
                        ref = quoted_ref(fn)

                    return self.store_value(fn.parent, "local", fn.original_name, ref)
            else:
                return self._visitFunctionDefinition(node)
        finally:
            self.units.pop()

    def _visitFunctionDefinition(self, node):
        fn = node.unit

        # localnames(name, ...)
        # globalnames(name, ...)

        localnames = fn.all_locals() and [
            compiler.ast.CallFunc(
                special_name("localnames"),
                [special_name(name) for name in fn.all_locals().keys()]
                )
            ] or []

        globalnames = fn.globals and [
            compiler.ast.CallFunc(
                special_name("globalnames"),
                [special_name(name) for name in fn.globals]
                )
            ] or []

        # Process any local class or function definitions.

        current = self.current_definition
        definitions = self.process_definitions(node)
        self.current_definition = current

        defaults = [self.dispatch(n) for n in node.defaults]

        # NOTE: Any required defaults should be copied into the local namespace
        # NOTE: using the __context__ reference to the instance of the function.

        # NOTE: Should generate guards for attribute usage operations.

        code = self.dispatch(node.code)

        return compiler.ast.Function(node.decorators, node.name, node.argnames, defaults, node.flags, node.doc,
            compiler.ast.Stmt(localnames + globalnames + definitions + code.nodes))

    visitGlobal = NOP

    def visitIf(self, node):
        return compiler.ast.If(
            [(self.dispatch(compare), self.dispatch(stmt)) for (compare, stmt) in node.tests],
            node.else_ and self.dispatch(node.else_)
            )

    def visitImport(self, node):

        # Generate __main__ function calls for each step in the imported module
        # hierarchy.

        statements = []

        for name, alias in node.names:
            for modname in self.module.get_module_paths(name):
                statements.append(
                    compiler.ast.CallFunc(compiler.ast.Getattr(modname, "__main__"), [])
                    )

            statements.append(
                compiler.ast.CallFunc(
                    special_name("storelocal"),
                    [special_name(alias or name.split(".")[0]),
                    compiler.ast.CallFunc(
                        special_name("static"),
                        [special_name(name)]
                        )
                    ])
                )

        return compiler.ast.Stmt(statements)

    def visitPass(self, node):
        if not isinstance(self.get_unit(), Class):
            return compiler.ast.Pass()
        else:
            return compiler.ast.Stmt([])

    def visitPrint(self, node):
        return compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("__builtins__", "_print"),
                node.dest and self.dispatch(node.dest) or special_name("None")]
                + [self.dispatch(n) for n in node.nodes]
            )

    def visitPrintnl(self, node):
        return compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("__builtins__", "_println"),
                node.dest and self.dispatch(node.dest) or special_name("None")]
                + [self.dispatch(n) for n in node.nodes]
            )

    def visitRaise(self, node):
        return compiler.ast.Raise(
            node.expr1 and self.dispatch(node.expr1),
            node.expr2 and self.dispatch(node.expr2),
            node.expr3 and self.dispatch(node.expr3)
            )

    def visitReturn(self, node):
        return compiler.ast.Return(self.dispatch(node.value))

    def visitStmt(self, node):
        return compiler.ast.Stmt([self.dispatch(n) for n in node.nodes])

    def visitTryExcept(self, node):
        # NOTE: Need to dispatch to the assignment with the exception.
        return compiler.ast.TryExcept(
            self.dispatch(node.body),
            [(spec and self.dispatch(spec), assign and self.dispatch(assign), self.dispatch(statement))
                for spec, assign, statement in node.handlers],
            node.else_ and self.dispatch(node.else_)
            )

    def visitTryFinally(self, node):
        return compiler.ast.TryFinally(
            self.dispatch(node.body),
            self.dispatch(node.final)
            )

    def visitWhile(self, node):
        return compiler.ast.While(
            self.dispatch(node.test),
            self.dispatch(node.body),
            node.else_ and self.dispatch(node.else_)
            )

    def visitYield(self, node):
        return compiler.ast.Yield(self.dispatch(node.value))

    # Expression-related helper methods.

    def _visitBitBinary(self, node):
        op_name = operator_functions[node.__class__.__name__]
        last = self.dispatch(node.nodes[0])

        for n in node.nodes[1:]:
            last = compiler.ast.CallFunc(
                special_name("apply"),
                [module_attribute("operator", op_name), last, self.dispatch(n)]
                )

        return last

    def _visitBinary(self, node):
        op_name = operator_functions[node.__class__.__name__]

        return compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("operator", op_name),
                self.dispatch(node.left), self.dispatch(node.right)]
            )

    def _visitUnary(self, node):
        op_name = operator_functions[node.__class__.__name__]

        return compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("operator", op_name), self.dispatch(node.expr)]
            )

    def _generateValue(self, value):

        # Literal constants.

        if isinstance(value, Const):
            return compiler.ast.Const(value.get_value())

        # Other constant structures.

        if isinstance(value, Constant):
            return quoted_ref(value)

        return None

    def _visitAttr(self, node, expr=None):
        unit = self.get_unit()

        # Choose the appropriate special functions.

        (opattrcontext, opattrcontextcond, opattr, opattrindexcontextcond, opconstant) = \
            expr and assattr_functions or getattr_functions

        accessor = self.dispatch(node.expr)

        # Generate already-deduced accesses.

        if node._access_type == "constant":
            value = self._generateValue(node._value_deduced)

            # Where constant attributes are accessed via instances, a special
            # operation setting the context is needed.

            if node._set_context == "set":
                return compiler.ast.CallFunc(special_name(opconstant), [value, accessor])
            else:
                return value

        # Generate accesses via static objects and instances.

        if node._attr_deduced:

            # Static attributes may cause context replacement.

            if node._access_type == "static":
                if node._set_context == "set":
                    op = opattrcontext
                elif node._set_context == "cond":
                    op = opattrcontextcond
                else:
                    op = opattr

                parent = self._generateValue(node._attr_deduced.parent)

            # Non-static attributes.

            else:
                op = opattr
                parent = None

            # Handle unsupported operations.

            if not op:
                raise TranslateError("Storing of class attribute %r via self not permitted." % node.attrname)

            # Define the arguments: accessor, attribute name and optional value.

            args = [
                parent or accessor,
                special_name(node.attrname)
                ]

            if expr:
                args.append(expr)

            # Append any context to be set.

            if node._set_context and args[0] is not accessor:
                args.append(accessor)

            return compiler.ast.CallFunc(special_name(op), args)

        # Positioned accesses are normal accesses via instances.

        if node._access_type == "positioned":
            args = [accessor, special_name(node.attrname)]
            if expr:
                args.append(expr)
            return compiler.ast.CallFunc(special_name(opattr), args)

        # With no usable deductions, generate a table-based access.

        args = [accessor, special_name(node.attrname)]
        if expr:
            args.append(expr)
        access = compiler.ast.CallFunc(special_name(opattrindexcontextcond), args)

        # class.__class__ => __builtins__.type

        if node.attrname == "__class__":

            # storetemp(n, <accessor>)
            # isclass(n) and __builtins__.type or <access>

            temp = quoted_name(unit.temp_usage)
            unit.temp_usage += 1

            return compiler.ast.Stmt([
                compiler.ast.CallFunc(special_name("storetemp"), [temp, access]),
                compiler.ast.Or([
                    compiler.ast.And([
                        compiler.ast.CallFunc(
                            special_name("isclass"),
                            [compiler.ast.CallFunc(special_name("loadtemp"), [temp])]
                            ),
                            module_attribute("__builtins__", "type")
                        ]),
                        access
                    ])
                ])

        # General accesses.

        else:
            return access

    # Expressions.

    def visitAdd(self, node):
        return self._visitBinary(node)

    def visitAnd(self, node):
        return compiler.ast.And([self.dispatch(n) for n in node.nodes])

    def visitAssAttr(self, node, expr=None):

        # Handle deletion.

        if compiler.ast.is_deletion(node):
            return compiler.ast.Stmt([])

        return self._visitAttr(node, expr)

    def visitAssList(self, node, expr=None):

        # Handle deletion.

        if compiler.ast.is_deletion(compiler.ast.flatten_assignment(node)):
            return compiler.ast.Stmt([])

        return compiler.ast.Stmt([
            self.dispatch(n, compiler.ast.CallFunc(
                special_name("apply"),
                [module_attribute("operator", "getitem"), expr, i]
                ))
            for (i, n) in enumerate(node.nodes)
            ])

    def visitAssName(self, node, expr=None):

        # Handle deletion.

        if compiler.ast.is_deletion(node):
            return compiler.ast.Stmt([])

        unit = self.get_unit()

        # Generate appropriate name access operation.
        # NOTE: Should generate guards for attribute usage operations.

        scope = getattr(node, "_scope", None)
        if not scope:
            attr, scope, from_name = self.get_unit()._get_with_scope(node.name)

        if scope == "constant":
            return node
        else:
            return self.store_value(unit, scope, node.name, expr)

    visitAssTuple = visitAssList

    def visitBitand(self, node):
        self._visitBitBinary(node)

    def visitBitor(self, node):
        self._visitBitBinary(node)

    def visitBitxor(self, node):
        self._visitBitBinary(node)

    def visitCallFunc(self, node):

        # Determine whether the invocation target is known.

        args = [self.dispatch(node.node)] + [self.dispatch(arg) for arg in node.args]

        target = node.node

        # Attribute information is only known for specific accessors.

        if isinstance(target, compiler.ast.AttributeAccessor):
            attr = target._attr
        else:
            attr = None

        if not attr or isinstance(attr, (Instance, UnresolvedName)):
            op = "apply"

        # Invocations with some knowledge available.

        else:
            context = attr.get_context()
            value = attr.get_value()

            # Class invocations.

            if isinstance(value, Class):
                op = "applyclass"
                if value:
                    args[0] = quoted_ref(value)
            else:

                # Function invocations.

                if context is ReplaceableContext:
                    op = "applyfunction"

                # Method invocations via classes.

                elif isinstance(context, Class) and value:
                    op = "applystaticmethod"
                    args[0] = quoted_ref(value)
                    args.insert(1, quoted_ref(context))

                # Plain method invocations.

                elif context:
                    op = "applymethod"

                # Unknown invocations.

                else:
                    op = "apply"

        return compiler.ast.CallFunc(
            special_name(op),
            args,
            node.star_args and self.dispatch(node.star_args),
            node.dstar_args and self.dispatch(node.dstar_args)
            )

    def visitCompare(self, node):
        nodes = []
        left = node.expr
        for op_name, right in node.ops:
            if op_name == "is":
                nodes.append(
                    compiler.ast.CallFunc(
                        special_name("__is__"),
                        [self.dispatch(left), self.dispatch(right)]
                        )
                    )
            elif op_name == "is not":
                nodes.append(
                    compiler.ast.CallFunc(
                        special_name("__is_not__"),
                        [self.dispatch(left), self.dispatch(right)]
                        )
                    )
            else:
                nodes.append(
                    compiler.ast.CallFunc(
                        special_name("apply"),
                        [module_attribute("operator", operator_functions.get(op_name)),
                            self.dispatch(left), self.dispatch(right)]
                        )
                    )
            left = right
        return compiler.ast.And(nodes)

    visitConst = NOP

    def visitDict(self, node):
        return compiler.ast.Dict([(self.dispatch(key), self.dispatch(value)) for (key, value) in node.items])

    def visitDiv(self, node):
        return self._visitBinary(node)

    def visitExec(self, node):

        # NOTE: Return the statement for now.

        return node

    def visitFloorDiv(self, node):
        return self._visitBinary(node)

    def visitGetattr(self, node, expr=None):
        return self._visitAttr(node, expr)

    def visitGenExpr(self, node):
        return compiler.ast.GenExpr(self.dispatch(node.code))

    def visitGenExprFor(self, node):
        expr = self.dispatch(node.iter)
        return compiler.ast.GenExprFor(
            self.dispatch(node.assign, expr), # NOTE: Needs to dispatch to AssName/AssTuple/AssList with an expression.
            expr,
            [self.dispatch(n) for n in node.ifs]
            )

    def visitGenExprIf(self, node):
        return compiler.ast.GenExprIf(self.dispatch(node.test))

    def visitGenExprInner(self, node):
        return compiler.ast.GenExprInner(
            self.dispatch(node.expr),
            [self.dispatch(n) for n in node.quals]
            )

    def visitIfExp(self, node):
        return compiler.ast.IfExp(
            self.dispatch(node.then),
            self.dispatch(node.test),
            self.dispatch(node.else_)
            )

    def visitInvert(self, node):
        return self._visitUnary(node)

    def visitKeyword(self, node):
        return compiler.ast.Keyword(
            node.name,
            self.dispatch(node.expr)
            )

    def visitLambda(self, node):
        self.units.append(node.unit)

        # NOTE: Need to initialise any defaults. Lambdas should probably be
        # NOTE: expanded to be "real" function definitions.

        try:
            return compiler.ast.Lambda(
                node.argnames,
                [self.dispatch(n) for n in node.defaults],
                node.flags,
                self.dispatch(node.code)
                )
        finally:
            self.units.pop()

    def visitLeftShift(self, node):
        return self._visitBinary(node)

    def visitList(self, node, expr=None):
        if expr:
            return self.visitAssList(node, expr)
        return compiler.ast.List([self.dispatch(n) for n in node.nodes])

    def visitListComp(self, node):

        """
        Convert from...

        [<expr> for <assign> in <list> [ for <assign> in <list> ]... [ if <test> ]... ]

        ...to...

        _out = []
        ...
        """

        unit = self.get_unit()
        temp = quoted_name(unit.temp_usage)
        unit.temp_usage += 1

        return compiler.ast.Stmt([
            # storetemp(_out, __builtins__.list())
            compiler.ast.CallFunc(special_name("storetemp"), [
                temp,
                compiler.ast.CallFunc(
                    special_name("apply"),
                    [module_attribute("__builtins__", "list")]
                    )
                ]),
            # ...
            self.dispatch(node.quals[0], temp, node.expr, node.quals[1:])
            ])

    def visitListCompFor(self, node, out_temp, expr, quals):

        """
        Convert from...

        [<expr> for <assign> in <list> ...]

        ...to...

        _it = iter(<list>)
        while True:
            try:
                <assign> = _it.next()
            except StopIteration:
                break
            else:
                ...
                _out.append(<expr>)
        """

        unit = self.get_unit()
        temp = quoted_name(unit.temp_usage)
        unit.temp_usage += 1

        # Either generate more "for" or "if" constructs.

        if node.ifs or quals:
            nodes = node.ifs + quals
            body = self.dispatch(nodes[0], out_temp, expr, nodes[1:])

        # Or generate the append statement.

        else:
            body = self._visitListCompExpr(out_temp, expr)

        # Wrap the above body in the loop construct.

        return compiler.ast.Stmt([
            # storetemp(_it, __builtins__.iter(<list>))
            compiler.ast.CallFunc(special_name("storetemp"), [
                temp,
                compiler.ast.CallFunc(
                    special_name("apply"),
                    [module_attribute("__builtins__", "iter"), self.dispatch(node.list)]
                    )
                ]),
            # while True: ...
            compiler.ast.While(
                special_name("True"),
                # try: ...
                compiler.ast.TryExcept(
                    compiler.ast.Stmt([
                        # <assign> = ...
                        self.dispatch(node.assign,
                            # _it.next()
                            compiler.ast.CallFunc(
                                compiler.ast.CallFunc(special_name("loadattrindex"), [
                                    compiler.ast.CallFunc(special_name("loadtemp"), [temp]),
                                    special_name("next")
                                    ]),
                                []
                                )
                            )
                        ]),
                    # except StopIteration: ...
                    [(special_name("StopIteration"), None, compiler.ast.Stmt([compiler.ast.Break()]))],
                    # else: ...
                    body
                    ),
                None
                )
            ])

    def visitListCompIf(self, node, out_temp, expr, quals):

        # Either generate more "for" or "if" constructs.

        if quals:
            body = self.dispatch(quals[0], out_temp, expr, quals[1:])

        # Or generate the append statement.

        else:
            body = self._visitListCompExpr(out_temp, expr)

        return compiler.ast.If(
            [(self.dispatch(node.test), body)],
            None
            )

    def _visitListCompExpr(self, out_temp, expr):

        "To the 'out_temp' object, append the result of 'expr'."

        return compiler.ast.Stmt([
            # _out.append(<expr>)
            compiler.ast.CallFunc(
                compiler.ast.CallFunc(special_name("loadattrindex"), [
                    compiler.ast.CallFunc(special_name("loadtemp"), [out_temp]),
                    special_name("append")
                    ]),
                [self.dispatch(expr)]
                )
            ])

    def visitMod(self, node):
        return self._visitBinary(node)

    def visitMul(self, node):
        return self._visitBinary(node)

    def visitName(self, node, expr=None):
        if expr:
            return self.visitAssName(node, expr)

        unit = self.get_unit()
        attr = node._attr
        scope = node._scope

        # Generate appropriate name access operation.

        if scope == "constant":
            return node

        # Function locals are referenced normally.

        elif scope == "local" and isinstance(unit, Function):
            return node

        # Other attributes should already be resolved.

        elif attr is not None and not isinstance(attr, Instance):
            if attr.is_constant():
                return constant_attribute(attr.parent, node.name)
            else:
                return compiler.ast.CallFunc(
                    special_name("loadattr"),
                    [quoted_ref(attr.parent), special_name(node.name)]
                    )

        # Function globals are referenced via the module.

        elif scope == "global":
            return compiler.ast.CallFunc(
                special_name("loadattr"),
                [quoted_ref(self.get_module()), special_name(node.name)]
                )

        # NOTE: Unknown attributes may arise because a class attribute has been
        # NOTE: optimised away.

        else:
            return compiler.ast.CallFunc(
                special_name("loadunknown"),
                [special_name(node.name)]
                )

    def visitNot(self, node):
        return compiler.ast.Not(self.dispatch(node.expr))

    def visitOr(self, node):
        return compiler.ast.Or([self.dispatch(n) for n in node.nodes])

    def visitPower(self, node):
        return self._visitBinary(node)

    def visitRightShift(self, node):
        return self._visitBinary(node)

    def visitSet(self, node):
        return compiler.ast.Set([self.dispatch(n) for n in node.nodes])

    def visitSlice(self, node, expr=None):
        return compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("operator", expr and "setslice" or "getslice"),
                self.dispatch(node.expr),
                node.lower and self.dispatch(node.lower),
                node.upper and self.dispatch(node.upper)]
                + (expr and [expr] or [])
            )

    def visitSliceobj(self, node):
        return compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("__builtins__", "slice")]
                + [self.dispatch(n) for n in node.nodes]
            )

    def visitSub(self, node):
        return self._visitBinary(node)

    def visitSubscript(self, node, expr=None):
        return compiler.ast.CallFunc(
            special_name("apply"),
            [module_attribute("operator", expr and "setitem" or "getitem"),
                self.dispatch(node.expr),
                compiler.ast.Tuple([self.dispatch(sub) for sub in node.subs])]
                + (expr and [expr] or [])
            )

    def visitTuple(self, node, expr=None):
        if expr:
            return self.visitAssTuple(node, expr)
        return compiler.ast.Tuple([self.dispatch(n) for n in node.nodes])

    def visitUnaryAdd(self, node):
        return self._visitUnary(node)

    def visitUnarySub(self, node):
        return self._visitUnary(node)

    def visitWith(self, node):

        """
        Convert from...

        with <expr> as <var>:
            ...

        ...to...

        _manager = <var> = <expr> # <var> may be absent
        _exit = _manager.__exit__
        _manager.__enter__()
        try:
            ...
        except Exception, exc:
            if not _exit(exc.type, exc.value, exc.tb):
                raise
        else:
            _exit(None, None, None)
        """

        # NOTE: For now, not adding this exuberance to the output.

        return node
            
# Convenience functions.

def convert(module, program, filename):
    stream = open(filename, "wb")
    try:
        source = ConvertedSource(module, program)
        source.to_stream(stream)
    finally:
        stream.close()

def translate(program, directory):
    if not exists(directory):
        os.mkdir(directory)

    # NOTE: Add constants here.

    for module in program.get_importer().get_modules():
        convert(module, program, join(directory, "%s%spy" % (module.full_name(), extsep)))

# vim: tabstop=4 expandtab shiftwidth=4
